@@ -8,13 +8,19 @@ from databuilder.models.neo4j_csv_serde import (
     RELATION_END_LABEL, RELATION_TYPE, RELATION_REVERSE_TYPE)
 from databuilder.publisher.neo4j_csv_publisher import UNQUOTED_SUFFIX
 
-DESCRIPTION_NODE_LABEL = 'Description'
+
+DESCRIPTION_NODE_LABEL_VAL = 'Description'
+DESCRIPTION_NODE_LABEL = DESCRIPTION_NODE_LABEL_VAL
 
 
-class TagMetadata:
+class TagMetadata(Neo4jCsvSerializable):
     TAG_NODE_LABEL = 'Tag'
     TAG_KEY_FORMAT = '{tag}'
     TAG_TYPE = 'tag_type'
+    DEFAULT_TYPE = 'default'
+    BADGE_TYPE = 'badge'
+    DASHBOARD_TYPE = 'dashboard'
+    METRIC_TYPE = 'metric'
 
     def __init__(self,
                  name,  # type: str,
@@ -22,6 +28,8 @@ class TagMetadata:
                  ):
         self._name = name
         self._tag_type = tag_type
+        self._nodes = iter([self.create_tag_node(self._name, self._tag_type)])
+        self._relations = iter([])
 
     @staticmethod
     def get_tag_key(name):
@@ -30,19 +38,111 @@ class TagMetadata:
             return ''
         return TagMetadata.TAG_KEY_FORMAT.format(tag=name)
 
+    @staticmethod
+    def create_tag_node(name, tag_type=DEFAULT_TYPE):
+        return {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
+                NODE_KEY: TagMetadata.get_tag_key(name),
+                TagMetadata.TAG_TYPE: tag_type}
+
+    def create_next_node(self):
+        # type: (...) -> Union[Dict[str, Any], None]
+        # return the string representation of the data
+        try:
+            return next(self._nodes)
+        except StopIteration:
+            return None
+
+    def create_next_relation(self):
+        # type: () -> Union[Dict[str, Any], None]
+        # We don't emit any relations for Tag ingestion
+        try:
+            return next(self._relations)
+        except StopIteration:
+            return None
+
+
+class DescriptionMetadata:
+    DESCRIPTION_NODE_LABEL = DESCRIPTION_NODE_LABEL_VAL
+    PROGRAMMATIC_DESCRIPTION_NODE_LABEL = 'Programmatic_Description'
+    DESCRIPTION_KEY_FORMAT = '{description}'
+    DESCRIPTION_TEXT = 'description'
+    DESCRIPTION_SOURCE = 'description_source'
+
+    DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
+    INVERSE_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION_OF'
+
+    # The default editable source.
+    DEFAULT_SOURCE = "description"
+
+    def __init__(self,
+                 text,  # type: Union[None, str]
+                 source=DEFAULT_SOURCE  # type: str
+                 ):
+        """
+        :param source: The unique source of what is populating this description.
+        :param text: the description text. Markdown supported.
+        """
+        self._source = source
+        self._text = text
+        #  There are so many dependencies on Description node, that it is probably easier to just separate the rest out.
+        if (self._source == self.DEFAULT_SOURCE):
+            self._label = self.DESCRIPTION_NODE_LABEL
+        else:
+            self._label = self.PROGRAMMATIC_DESCRIPTION_NODE_LABEL
+
+    @staticmethod
+    def create_description_metadata(text, source=DEFAULT_SOURCE):
+        # type: (Union[None,str], str) -> ProgrammaticDescription
+
+        # We do not want to create a node if there is no description text!
+        if text is None:
+            return None
+        if not source:
+            description_node = DescriptionMetadata(text=text, source=DescriptionMetadata.DEFAULT_SOURCE)
+        else:
+            description_node = DescriptionMetadata(text=text, source=source)
+        return description_node
+
+    def get_description_id(self):
+        # type: () -> str
+        if self._source == self.DEFAULT_SOURCE:
+            return "_description"
+        else:
+            return "_" + self._source + "_description"
+
+    def __repr__(self):
+        # type: () -> str
+        return 'DescriptionMetadata({!r}, {!r})'.format(self._source, self._text)
+
+    def get_node_dict(self, node_key):
+        # (str) -> Dict
+        return {
+            NODE_LABEL: self._label,
+            NODE_KEY: node_key,
+            DescriptionMetadata.DESCRIPTION_SOURCE: self._source,
+            DescriptionMetadata.DESCRIPTION_TEXT: self._text,
+        }
+
+    def get_relation(self, start_node, start_key, end_key):
+        # (str, str) => Dict
+        return {
+            RELATION_START_LABEL: start_node,
+            RELATION_END_LABEL: self._label,
+            RELATION_START_KEY: start_key,
+            RELATION_END_KEY: end_key,
+            RELATION_TYPE: DescriptionMetadata.DESCRIPTION_RELATION_TYPE,
+            RELATION_REVERSE_TYPE: DescriptionMetadata.INVERSE_DESCRIPTION_RELATION_TYPE
+        }
+
 
 class ColumnMetadata:
     COLUMN_NODE_LABEL = 'Column'
     COLUMN_KEY_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}'
     COLUMN_NAME = 'name'
     COLUMN_TYPE = 'type'
-    COLUMN_ORDER = 'sort_order'
+    COLUMN_ORDER = 'sort_order{}'.format(UNQUOTED_SUFFIX)  # int value needs to be unquoted when publish to neo4j
     COLUMN_DESCRIPTION = 'description'
-    COLUMN_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}/_description'
-
-    # pair of nodes makes relationship where name of variable represents order of relationship.
-    COL_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
-    DESCRIPTION_COL_RELATION_TYPE = 'DESCRIPTION_OF'
+    COLUMN_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{col}/{description_id}'
 
     # Relation between column and tag
     COL_TAG_RELATION_TYPE = 'TAGGED_BY'
@@ -53,7 +153,7 @@ class ColumnMetadata:
                  description,  # type: Union[str, None]
                  col_type,  # type: str
                  sort_order,  # type: int
-                 tags=None,  # Union[List[str], None]
+                 tags=None  # type: Union[List[str], None]
                  ):
         # type: (...) -> None
         """
@@ -64,7 +164,8 @@ class ColumnMetadata:
         :param sort_order:
         """
         self.name = name
-        self.description = description
+        self.description = DescriptionMetadata.create_description_metadata(source=None,
+                                                                           text=description)
         self.type = col_type
         self.sort_order = sort_order
         self.tags = tags
@@ -98,10 +199,7 @@ class TableMetadata(Neo4jCsvSerializable):
     TABLE_NAME = 'name'
     IS_VIEW = 'is_view{}'.format(UNQUOTED_SUFFIX)  # bool value needs to be unquoted when publish to neo4j
 
-    TABLE_DESCRIPTION = 'description'
-    TABLE_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/_description'
-    TABLE_DESCRIPTION_RELATION_TYPE = 'DESCRIPTION'
-    DESCRIPTION_TABLE_RELATION_TYPE = 'DESCRIPTION_OF'
+    TABLE_DESCRIPTION_FORMAT = '{db}://{cluster}.{schema}/{tbl}/{description_id}'
 
     DATABASE_NODE_LABEL = 'Database'
     DATABASE_KEY_FORMAT = 'database://{db}'
@@ -131,36 +229,38 @@ class TableMetadata(Neo4jCsvSerializable):
     def __init__(self,
                  database,  # type: str
                  cluster,  # type: str
-                 schema_name,  # type: str
+                 schema,  # type: str
                  name,  # type: str
                  description,  # type: Union[str, None]
                  columns=None,  # type: Iterable[ColumnMetadata]
                  is_view=False,  # type: bool
                  tags=None,  # type: Union[List, str]
+                 description_source=None,  # type: Union[str, None]
                  **kwargs  # type: Dict
                  ):
         # type: (...) -> None
         """
         :param database:
         :param cluster:
-        :param schema_name:
+        :param schema:
         :param name:
         :param description:
         :param columns:
         :param is_view: Indicate whether the table is a view or not
         :param tags:
+        :param description_source: Optional. Where the description is coming from. Used to compose unique id.
         :param kwargs: Put additional attributes to the table model if there is any.
         """
         self.database = database
         self.cluster = cluster
-        self.schema_name = schema_name
+        self.schema = schema
         self.name = name
-        self.description = description
+        self.description = DescriptionMetadata.create_description_metadata(text=description, source=description_source)
         self.columns = columns if columns else []
         self.is_view = is_view
         self.attrs = None
         if isinstance(tags, str):
-            tags = tags.split(',')
+            tags = list(filter(None, tags.split(',')))
         if isinstance(tags, list):
             tags = [tag.lower().strip() for tag in tags]
         self.tags = tags
@@ -176,7 +276,7 @@ class TableMetadata(Neo4jCsvSerializable):
         return 'TableMetadata({!r}, {!r}, {!r}, {!r} ' \
             '{!r}, {!r}, {!r}, {!r})'.format(self.database,
                                              self.cluster,
-                                             self.schema_name,
+                                             self.schema,
                                              self.name,
                                              self.description,
                                              self.columns,
@@ -187,15 +287,16 @@ class TableMetadata(Neo4jCsvSerializable):
         # type: () -> str
         return TableMetadata.TABLE_KEY_FORMAT.format(db=self.database,
                                                      cluster=self.cluster,
-                                                     schema=self.schema_name,
+                                                     schema=self.schema,
                                                      tbl=self.name)
 
-    def _get_table_description_key(self):
-        # type: () -> str
+    def _get_table_description_key(self, description):
+        # type: (DescriptionMetadata) -> str
         return TableMetadata.TABLE_DESCRIPTION_FORMAT.format(db=self.database,
                                                              cluster=self.cluster,
-                                                             schema=self.schema_name,
-                                                             tbl=self.name)
+                                                             schema=self.schema,
+                                                             tbl=self.name,
+                                                             description_id=description.get_description_id())
 
     def _get_database_key(self):
         # type: () -> str
@@ -210,23 +311,24 @@ class TableMetadata(Neo4jCsvSerializable):
         # type: () -> str
         return TableMetadata.SCHEMA_KEY_FORMAT.format(db=self.database,
                                                       cluster=self.cluster,
-                                                      schema=self.schema_name)
+                                                      schema=self.schema)
 
     def _get_col_key(self, col):
         # type: (ColumnMetadata) -> str
         return ColumnMetadata.COLUMN_KEY_FORMAT.format(db=self.database,
                                                        cluster=self.cluster,
-                                                       schema=self.schema_name,
+                                                       schema=self.schema,
                                                        tbl=self.name,
                                                        col=col.name)
 
-    def _get_col_description_key(self, col):
-        # type: (ColumnMetadata) -> str
+    def _get_col_description_key(self, col, description):
+        # type: (ColumnMetadata, DescriptionMetadata) -> str
         return ColumnMetadata.COLUMN_DESCRIPTION_FORMAT.format(db=self.database,
                                                                cluster=self.cluster,
-                                                               schema=self.schema_name,
+                                                               schema=self.schema,
                                                                tbl=self.name,
-                                                               col=col.name)
+                                                               col=col.name,
+                                                               description_id=description.get_description_id())
 
     def create_next_node(self):
         # type: () -> Union[Dict[str, Any], None]
@@ -249,16 +351,13 @@ class TableMetadata(Neo4jCsvSerializable):
         yield table_node
 
         if self.description:
-            yield {NODE_LABEL: DESCRIPTION_NODE_LABEL,
-                   NODE_KEY: self._get_table_description_key(),
-                   TableMetadata.TABLE_DESCRIPTION: self.description}
+            node_key = self._get_table_description_key(self.description)
+            yield self.description.get_node_dict(node_key)
 
         # Create the table tag node
         if self.tags:
             for tag in self.tags:
-                yield {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
-                       NODE_KEY: TagMetadata.get_tag_key(tag),
-                       TagMetadata.TAG_TYPE: 'default'}
+                yield TagMetadata.create_tag_node(tag)
 
         for col in self.columns:
             yield {
@@ -268,21 +367,15 @@ class TableMetadata(Neo4jCsvSerializable):
                 ColumnMetadata.COLUMN_TYPE: col.type,
                 ColumnMetadata.COLUMN_ORDER: col.sort_order}
 
-            if not col.description:
-                continue
+            if col.description:
+                node_key = self._get_col_description_key(col, col.description)
+                yield col.description.get_node_dict(node_key)
 
-            yield {
-                NODE_LABEL: DESCRIPTION_NODE_LABEL,
-                NODE_KEY: self._get_col_description_key(col),
-                ColumnMetadata.COLUMN_DESCRIPTION: col.description}
-
-            if not col.tags:
-                continue
-
-            for tag in col.tags:
-                yield {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
-                       NODE_KEY: TagMetadata.get_tag_key(tag),
-                       TagMetadata.TAG_TYPE: 'default'}
+            if col.tags:
+                for tag in col.tags:
+                    yield {NODE_LABEL: TagMetadata.TAG_NODE_LABEL,
+                           NODE_KEY: TagMetadata.get_tag_key(tag),
+                           TagMetadata.TAG_TYPE: 'default'}
 
         # Database, cluster, schema
         others = [NodeTuple(key=self._get_database_key(),
@@ -292,7 +385,7 @@ class TableMetadata(Neo4jCsvSerializable):
                             name=self.cluster,
                             label=TableMetadata.CLUSTER_NODE_LABEL),
                   NodeTuple(key=self._get_schema_key(),
-                            name=self.schema_name,
+                            name=self.schema,
                             label=TableMetadata.SCHEMA_NODE_LABEL)
                   ]
 
@@ -325,14 +418,9 @@ class TableMetadata(Neo4jCsvSerializable):
         }
 
         if self.description:
-            yield {
-                RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
-                RELATION_END_LABEL: DESCRIPTION_NODE_LABEL,
-                RELATION_START_KEY: self._get_table_key(),
-                RELATION_END_KEY: self._get_table_description_key(),
-                RELATION_TYPE: TableMetadata.TABLE_DESCRIPTION_RELATION_TYPE,
-                RELATION_REVERSE_TYPE: TableMetadata.DESCRIPTION_TABLE_RELATION_TYPE
-            }
+            yield self.description.get_relation(TableMetadata.TABLE_NODE_LABEL,
+                                                self._get_table_key(),
+                                                self._get_table_description_key(self.description))
 
         if self.tags:
             for tag in self.tags:
@@ -355,30 +443,21 @@ class TableMetadata(Neo4jCsvSerializable):
                 RELATION_REVERSE_TYPE: TableMetadata.COL_TABLE_RELATION_TYPE
             }
 
-            if not col.description:
-                continue
+            if col.description:
+                yield col.description.get_relation(ColumnMetadata.COLUMN_NODE_LABEL,
+                                                   self._get_col_key(col),
+                                                   self._get_col_description_key(col, col.description))
 
-            yield {
-                RELATION_START_LABEL: ColumnMetadata.COLUMN_NODE_LABEL,
-                RELATION_END_LABEL: DESCRIPTION_NODE_LABEL,
-                RELATION_START_KEY: self._get_col_key(col),
-                RELATION_END_KEY: self._get_col_description_key(col),
-                RELATION_TYPE: ColumnMetadata.COL_DESCRIPTION_RELATION_TYPE,
-                RELATION_REVERSE_TYPE: ColumnMetadata.DESCRIPTION_COL_RELATION_TYPE
-            }
-
-            if not col.tags:
-                continue
-
-            for tag in col.tags:
-                yield {
-                    RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
-                    RELATION_END_LABEL: TagMetadata.TAG_NODE_LABEL,
-                    RELATION_START_KEY: self._get_table_key(),
-                    RELATION_END_KEY: TagMetadata.get_tag_key(tag),
-                    RELATION_TYPE: ColumnMetadata.COL_TAG_RELATION_TYPE,
-                    RELATION_REVERSE_TYPE: ColumnMetadata.TAG_COL_RELATION_TYPE,
-                }
+            if col.tags:
+                for tag in col.tags:
+                    yield {
+                        RELATION_START_LABEL: TableMetadata.TABLE_NODE_LABEL,
+                        RELATION_END_LABEL: TagMetadata.TAG_NODE_LABEL,
+                        RELATION_START_KEY: self._get_table_key(),
+                        RELATION_END_KEY: TagMetadata.get_tag_key(tag),
+                        RELATION_TYPE: ColumnMetadata.COL_TAG_RELATION_TYPE,
+                        RELATION_REVERSE_TYPE: ColumnMetadata.TAG_COL_RELATION_TYPE,
+                    }
 
         others = [
             RelTuple(start_label=TableMetadata.DATABASE_NODE_LABEL,
